"""
Set function annotations manually.
"""

from lib2to3 import fixer_base
from lib2to3.pgen2 import token
from lib2to3.pytree import Node, Leaf
from lib2to3.fixer_util import syms, Assign, Name, Attr, Comma, String, Newline

from ..fixer_util import indentation

from functools import partial
from lib3to2.fixer_util import paste_after

_LBrace = partial(Leaf, token.LBRACE, '{')
_RBrace = partial(Leaf, token.RBRACE, '}')
_Colon = partial(Leaf, token.COLON, ':')

def _build_pair(k, v):
    v.prefix = ' '
    return [k, _Colon(), v]

def _commatize(pairs):
    prefix = None
    for pair in pairs:
        for item in pair:
            if prefix:
                item.prefix = prefix
                prefix = None
            yield item
        yield Comma()
        prefix = ' '

def _build_pair_list(items):
    return list(_commatize(_build_pair(k, v) for k, v in items))[:-1]

def build_dict(items):
    pair_list = _build_pair_list(items)
    if pair_list == []:
        return Node(syms.atom, [_LBrace(), _RBrace()])
    else:
        body = Node(syms.dictsetmaker, pair_list)
        return Node(syms.atom, [_LBrace(), body, _RBrace()])

class FixAnnotations(fixer_base.BaseFix):

    order = "pre"

    PATTERN = """
              funcdef< 'def' name=NAME parameters< '(' [params=any] ')' > ['->' ret=any] ':' any* >
              """

    def transform(self, node, results):
        name = results['name']
        params_node = results.get("params")
        ret = results.get("ret")
        
        annotations = []
        
        if params_node is not None:
            params = []
            if params_node.type == syms.typedargslist:
                params = params_node.children
            elif params_node.type == syms.tname:
                params.append(params_node)
                
            for param in params:
                if param.type == syms.tname:
                    tname, _, tvalue = param.children
                    annotations.append((String("'"+tname.value+"'"), 
                                        tvalue.clone()))
                    param.replace(tname)
        
        if ret is not None:
            assert ret.prev_sibling.type == token.RARROW, "Invalid return annotation"
            annotations.append((String("'return'"), ret.clone()))
            ret.prev_sibling.remove()
            ret.remove()
        
        
        annotations_attr = Attr(Name(name.value), Name('func_annotations'))
        annotations_assign = Assign(annotations_attr, build_dict(annotations))
        
        
        annotations_line = Node(syms.simple_stmt, 
                                [annotations_assign, Newline()])
        
        paste_after(node, annotations_line)