'''
Converts automatic numbering (Py3, Py2.7) to explicit numbering,
that is Py2.6 compatible.
'''

import string as _string

from itertools import chain as _chain

from lib2to3 import fixer_base, patcomp
from lib2to3.pgen2 import token
from lib2to3.pygram import python_symbols as syms
from lib2to3.fixer_util import String

from .fix_str import FixStr as _FixStr

_digits = frozenset(_string.digits) 

_formatter = _string.Formatter()

def _parse_formatstring(format_string):
    return _formatter.parse(format_string)

class _FormatStringCounter(object):
    
    def __init__(self, start_idx=0):
        self._counter = start_idx
        self._idx_present = False
        
    def __int__(self):
        return self._counter
    
    def __str__(self):
        return str(int(self))
    
    def __repr__(self):
        return '{0}({1})'.format(type(self).__name__, int(self))
    
    def parse_name(self, name):
        if not name:
            if self._idx_present:
                _msg = ("cannot switch from manual field specification" 
                        "to automatic field numbering")
                raise ValueError(_msg)
            else:
                name = str(self._counter)
                self._counter += 1
        elif name[0] in _digits:
            self._idx_present = True
        
        return name

def _fix_formatstring(tmpl, counter):
    try:
        tmpl_iterator = _parse_formatstring(tmpl)
    except Exception as e:
        _msg = "Invalid format string: {0!r}".format(e)
        raise ValueError(_msg)
    
    try:
        for prefix, name, spec, conv in tmpl_iterator:
            yield prefix
            
            if name is spec is conv is None:
                if prefix == '{':
                    yield '{'
                elif prefix.endswith('}'):
                    yield '}'
            else:
                name = counter.parse_name(name)
                if spec and spec.startswith('{') and spec.endswith('}'):
                    spec = '{' + counter.parse_name(spec[1:-1]) + '}'
                
                result = '{' + name
                if conv:
                    result += '!'+conv
                if spec:
                    result += ':'+spec
                
                yield result + '}'
             
    except ValueError as e:
        _msg = "Invalid format string: {0!r}".format(e)
        raise ValueError(_msg)

def fix_formatstring(tmpl, start_idx):
    counter = _FormatStringCounter(start_idx)
    fixed_str = ''.join(_fix_formatstring(tmpl, counter))
    return int(counter), fixed_str

def parse_string_literal(literal):
    try:
        return eval(literal)
    except:
        _msg = 'Error while parsing string literal: {!r}'
        raise ValueError(_msg.format(literal))

def _get_string_literals(base):
    if base.type == token.STRING:
        return [base]
    
    if base.children:
        base = base.children
    
    if len(base) == 3 and (base[0].type == token.LPAR and
                           base[2].type == token.RPAR):
        base = base[1]
        if base.type == syms.atom:
            base = base.children
        elif base.type == token.STRING:
            return [base]
    
    if all(l.type == token.STRING for l in base):
        return base
    return []

def _find_suite(node):
    while node.type not in (syms.suite, syms.file_input):
        node = node.parent
    return node

_simpl_binding_pattern = patcomp.compile_pattern(
  "expr_stmt< NAME '=' any >")
def _find_last_simple_binding_in_suite(name, node):
    node_lineno = node.get_lineno()
    preceding_stmts = (stmt for stmt in reversed(_find_suite(node).children)
                       if stmt.get_lineno() < node_lineno)
    simple_stmts = (stmt for stmt in preceding_stmts 
                    if stmt.type == syms.simple_stmt)
    small_stmts = _chain(*[simple_stmt.children for simple_stmt
                           in simple_stmts])
    binding_stmts = filter(_simpl_binding_pattern.match, small_stmts)
    return next((stmt.children[2] for stmt in binding_stmts
                 if stmt.children[0].value == name), None)

class FixFormatNumbering(fixer_base.BaseFix):
    
    order = 'pre'
    run_order = _FixStr.run_order - 1
    
    PATTERN = """
    power< base=any trailer< '.' 'format' > any* >"""
    
    def transform(self, node, results):

        base = results.get("base")
        
        if base:
            if base.type == token.NAME:
                binding = _find_last_simple_binding_in_suite(base.value, node)
                if binding:
                    self.__transform_tmpl(binding)
                else:
                    self.cannot_convert(node,
                      "Possibly formatting, but cannot match format string.")
            else:
                self.__transform_tmpl(base)
    
    def __transform_tmpl(self, base):
        cur_idx = 0
        for literal in _get_string_literals(base):
            formatstring = parse_string_literal(literal.value)
            cur_idx, fixed_str = fix_formatstring(formatstring, cur_idx)
            literal.replace(String(repr(fixed_str), literal.prefix))